/*************************************************************************

   LameCrypt 128-bit block cipher with 512 bit secret key

   Written in 2009 by Alexander Myasnikow

   Web: www.darksoftware.narod.ru

*************************************************************************/

#include <stdlib.h>
#include <string.h>

typedef unsigned char u8;
typedef unsigned long u32;
typedef signed long s32;

#define rol(x,r) _lrotl(x,r)

#include "sbox.h"

#define CNT_INDEX 0

u8 extable[32][256];

u32 fulltable[32][256];

u32
subst (u32 x)
{
  x =
    rol (k87[x >> 24 & 255] << 24 | k65[x >> 16 & 255] << 16 |
	 k43[x >> 8 & 255] << 8 | k21[x & 255], 11);

  return x;
}

u32 rawkey[16];

u32 xrec[32][4];



void __stdcall 
setup (u8 * key)
{
  u32 i, n, c = 0, counter = 0, r, x = CNT_INDEX;

  u32 *ikey = (u32 *) & key[0], *skey = (u32 *) & key[32];

  for (i = 0; i != 32; i++)

    for (n = 0; n != 256; n++)
      extable[i][n] = exmat[key[i] ^ n];



  for (i = 0; i != 32; i++)
    {
      for (n = 0; n != 256; n++)
	{
	  counter += 0x9e3779b9;

	  fulltable[i][n] =
	    subst (rol
		   (keymat[i % 8][extable[i][n]] ^
		    rol (ikey[n % 8] + counter, n % 32),
		    i) + rol (skey[c % 8], rol (counter, c & 31) >> 27));

	  c += 1;
	}

      for (r = 0; r != 4; r++)

	xrec[i][r] = x++;
    }


  for (i = 0; i != 8; i++)

    rawkey[i] = subst (ikey[i] + skey[i]);
}


void __stdcall 
crypt (u32 * plain)
{
  u32 a, b, c, d, r;

  a = plain[0] + rawkey[0];
  b = plain[1] + rawkey[1];
  c = plain[2] + rawkey[2];
  d = plain[3] + rawkey[3];

  for (r = 0; r != 32; r++)
    {
      a +=
	(b + (b << 6) ^ (b >> 8)) +
	fulltable[r][(xrec[r][0] + rol (b, c >> 27)) & 255];
      b +=
	(c + (c << 6) ^ (c >> 8)) +
	fulltable[r][(xrec[r][1] + rol (c, d >> 27)) & 255];
      c +=
	(d + (d << 6) ^ (d >> 8)) +
	fulltable[r][(xrec[r][2] + rol (d, a >> 27)) & 255];
      d +=
	(a + (a << 6) ^ (a >> 8)) +
	fulltable[r][(xrec[r][3] + rol (a, b >> 27)) & 255];
    }

  plain[0] = a ^ rawkey[4];
  plain[1] = b ^ rawkey[5];
  plain[2] = c ^ rawkey[6];
  plain[3] = d ^ rawkey[7];
}


void __stdcall 
decrypt (u32 * plain)
{
  u32 a, b, c, d;
  s32 r;

  a = plain[0] ^ rawkey[4];
  b = plain[1] ^ rawkey[5];
  c = plain[2] ^ rawkey[6];
  d = plain[3] ^ rawkey[7];

  for (r = 31; r != -1; r--)
    {
      d -=
	(a + (a << 6) ^ (a >> 8)) +
	fulltable[r][(xrec[r][3] + rol (a, b >> 27)) & 255];
      c -=
	(d + (d << 6) ^ (d >> 8)) +
	fulltable[r][(xrec[r][2] + rol (d, a >> 27)) & 255];
      b -=
	(c + (c << 6) ^ (c >> 8)) +
	fulltable[r][(xrec[r][1] + rol (c, d >> 27)) & 255];
      a -=
	(b + (b << 6) ^ (b >> 8)) +
	fulltable[r][(xrec[r][0] + rol (b, c >> 27)) & 255];
    }

  plain[0] = a - rawkey[0];
  plain[1] = b - rawkey[1];
  plain[2] = c - rawkey[2];
  plain[3] = d - rawkey[3];
}



u32 __stdcall 
getblocksize ()
{
  return 128;
}

u32 __stdcall 
getkeysize ()
{
  return 512;
}

void __stdcall 
getciphername (u8 * p)
{
  strcpy (p, "LameCrypt");
}
